"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.KafkaNodeProducer = void 0;
const events_1 = require("events");
const util = __importStar(require("util"));
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_client_1 = require("@fluidframework/server-services-client");
const kafka = __importStar(require("kafka-node"));
const kafkaTopics_1 = require("./kafkaTopics");
/**
 * Kafka producer using the kafka-node library
 */
class KafkaNodeProducer {
    constructor(clientOptions, clientId, topic, topicPartitions, topicReplicationFactor, maxBatchSize, maxMessageSize) {
        this.clientOptions = clientOptions;
        this.topic = topic;
        this.topicPartitions = topicPartitions;
        this.topicReplicationFactor = topicReplicationFactor;
        this.messages = new Map();
        this.events = new events_1.EventEmitter();
        this.connecting = false;
        this.connected = false;
        clientOptions.clientId = clientId;
        this.maxBatchSize = maxBatchSize ?? server_services_core_1.MaxBatchSize;
        this.maxMessageSize = maxMessageSize ?? Number.MAX_SAFE_INTEGER;
        this.connect();
    }
    isConnected() {
        return this.connected;
    }
    /**
     * Sends the provided message to Kafka
     */
    // eslint-disable-next-line @typescript-eslint/promise-function-async
    send(messages, tenantId, documentId) {
        const key = `${tenantId}/${documentId}`;
        // Get the list of boxcars for the given key
        if (!this.messages.has(key)) {
            this.messages.set(key, [new server_services_core_1.PendingBoxcar(tenantId, documentId)]);
        }
        const boxcars = this.messages.get(key);
        // Create a new boxcar if necessary (will only happen when not connected)
        if (boxcars[boxcars.length - 1].messages.length + messages.length >= this.maxBatchSize) {
            boxcars.push(new server_services_core_1.PendingBoxcar(tenantId, documentId));
        }
        // Add the message to the boxcar
        const boxcar = boxcars[boxcars.length - 1];
        boxcar.messages.push(...messages);
        // If adding a new message to the boxcar filled it up, and we are connected, then send immediately. Otherwise
        // request a send
        if (this.connected && boxcar.messages.length >= this.maxBatchSize) {
            // Send all the boxcars
            this.sendBoxcars(boxcars);
            this.messages.delete(key);
        }
        else {
            // Mark the need to send a message
            this.requestSend();
        }
        return boxcar.deferred.promise;
    }
    async close() {
        await util.promisify(((callback) => this.producer.close(callback)))();
        await util.promisify(((callback) => this.client.close(callback)))();
    }
    on(event, listener) {
        this.events.on(event, listener);
        return this;
    }
    once(event, listener) {
        this.events.once(event, listener);
        return this;
    }
    /**
     * Notifies of the need to send pending messages. We defer sending messages to batch together messages
     * to the same partition.
     */
    requestSend() {
        // If we aren't connected yet defer sending until connected
        if (!this.connected) {
            return;
        }
        // Exit early if there is a pending send
        if (this.sendPending) {
            return;
        }
        // Use setImmediate to play well with the node event loop
        this.sendPending = setImmediate(() => {
            this.sendPendingMessages();
            this.sendPending = undefined;
        });
    }
    /**
     * Sends all pending messages
     */
    sendPendingMessages() {
        for (const [, value] of this.messages) {
            this.sendBoxcars(value);
        }
        this.messages.clear();
    }
    sendBoxcars(boxcars) {
        for (const boxcar of boxcars) {
            const boxcarMessage = {
                contents: boxcar.messages,
                documentId: boxcar.documentId,
                tenantId: boxcar.tenantId,
                type: server_services_core_1.BoxcarType,
            };
            const stringifiedMessage = Buffer.from(JSON.stringify(boxcarMessage));
            if (stringifiedMessage.byteLength > this.maxMessageSize) {
                const error = new server_services_client_1.NetworkError(413, `Boxcar message size (${stringifiedMessage.byteLength}) exceeded max message size (${this.maxMessageSize})`);
                boxcar.deferred.reject(error);
                continue;
            }
            this.producer.send([{ key: boxcar.documentId, messages: stringifiedMessage, topic: this.topic }], (error, data) => {
                if (error) {
                    this.handleError(error);
                    boxcar.deferred.reject(error);
                }
                else {
                    this.events.emit("produced");
                    boxcar.deferred.resolve();
                }
            });
        }
    }
    /**
     * Creates a connection to Kafka. Will reconnect on failure.
     */
    connect() {
        // Exit out if we are already connected or are in the process of connecting
        if (this.connected || this.connecting) {
            return;
        }
        this.connecting = true;
        this.client = new kafka.KafkaClient(this.clientOptions);
        this.producer = new kafka.Producer(this.client, { partitionerType: 3 });
        this.client.on("error", (error) => {
            this.handleError(error);
        });
        this.producer.on("ready", async () => {
            try {
                await (0, kafkaTopics_1.ensureTopics)(this.client, [this.topic], this.topicPartitions, this.topicReplicationFactor);
                this.connected = true;
                this.connecting = false;
                this.events.emit("connected");
                this.sendPendingMessages();
            }
            catch (error) {
                this.handleError(error);
            }
        });
        this.producer.on("error", (error) => {
            this.handleError(error);
        });
    }
    /**
     * Handles an error that requires a reconnect to Kafka
     */
    handleError(error) {
        // Close the client if it exists
        if (this.client) {
            this.client.close();
            this.client = undefined;
        }
        this.connecting = this.connected = false;
        this.events.emit("error", error);
        this.connect();
    }
}
exports.KafkaNodeProducer = KafkaNodeProducer;
//# sourceMappingURL=kafkaNodeProducer.js.map